@@ -1,22 +1,9 @@
-/***************************************************************************
- *   This file is part of KDevelop                                         *
- *   Copyright 2008 Niko Sams <niko.sams@gmail.com>                        *
- *                                                                         *
- *   This program is free software; you can redistribute it and/or modify  *
- *   it under the terms of the GNU Library General Public License as       *
- *   published by the Free Software Foundation; either version 2 of the    *
- *   License, or (at your option) any later version.                       *
- *                                                                         *
- *   This program is distributed in the hope that it will be useful,       *
- *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
- *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
- *   GNU General Public License for more details.                          *
- *                                                                         *
- *   You should have received a copy of the GNU Library General Public     *
- *   License along with this program; if not, write to the                 *
- *   Free Software Foundation, Inc.,                                       *
- *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.         *
- ***************************************************************************/
+/*
+    SPDX-FileCopyrightText: 2008 Niko Sams <niko.sams@gmail.com>
+
+    SPDX-License-Identifier: LGPL-2.0-or-later
+*/
+
 #include "expressionvisitor.h"
 #include "parsesession.h"
 #include "editorintegrator.h"
@@ -194,12 +181,12 @@ void ExpressionVisitor::visitVarExpression(VarExpressionAst *node)
 {
     DefaultVisitor::visitVarExpression(node);
     if (node->isGenerator != -1) {
-        DeclarationPointer generatorDecl = findDeclarationImport(ClassDeclarationType, QualifiedIdentifier("generator"));
+        DeclarationPointer generatorDecl = findDeclarationImport(ClassDeclarationType, QualifiedIdentifier(u"generator"));
 
         if (generatorDecl) {
             m_result.setType(AbstractType::Ptr(new IntegralType(IntegralType::TypeMixed)));
             if (hasCurrentClosureReturnType()) {
-                FunctionType::Ptr closureType = currentClosureReturnType().cast<FunctionType>();
+                auto closureType = currentClosureReturnType().staticCast<FunctionType>();
                 closureType->setReturnType(generatorDecl->abstractType());
             }
         }
@@ -237,13 +224,8 @@ void ExpressionVisitor::visitClosure(ClosureAst* node)
     if (node->functionBody) {
         visitInnerStatementList(node->functionBody);
     }
-    if (node->returnType && node->returnType->typehint && isClassTypehint(node->returnType->typehint, m_editor)) {
-        NamespacedIdentifierAst* objectType = node->returnType->typehint->genericType;
-        QualifiedIdentifier id = identifierForNamespace(objectType, m_editor);
-        DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
-
-        usingDeclaration(objectType->namespaceNameSequence->back()->element, dec);
-        buildNamespaceUses(objectType, id);
+    if (node->returnType) {
+        visitReturnType(node->returnType);
     }
 
     //Override found type with return typehint or phpdoc return typehint
@@ -259,15 +241,10 @@ void ExpressionVisitor::visitClosure(ClosureAst* node)
             AbstractType::Ptr type = parameterType(it->element, {}, m_editor, m_currentContext);
             closureType->addArgument(type);
 
-            if (it->element->parameterType && it->element->parameterType->typehint
-                    && isClassTypehint(it->element->parameterType->typehint, m_editor)) {
-                NamespacedIdentifierAst* objectType = it->element->parameterType->typehint->genericType;
-                QualifiedIdentifier id = identifierForNamespace(objectType, m_editor);
-                DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
-
-                usingDeclaration(objectType->namespaceNameSequence->back()->element, dec);
-                buildNamespaceUses(objectType, id);
+            if (it->element->parameterType) {
+                visitParameterType(it->element->parameterType);
             }
+
             if (it->element->defaultValue) {
                 visitExpr(it->element->defaultValue);
             }
@@ -598,8 +575,7 @@ void ExpressionVisitor::visitVariableProperty(VariablePropertyAst *node)
 
         //If the variable type is unsure, try to see if it contains a StructureType. If so, use that
         //    (since the other types do not allow accessing properties)
-        if (type && type.cast<UnsureType>()) {
-            UnsureType::Ptr unsureType = type.cast<UnsureType>();
+        if (auto unsureType = type.dynamicCast<UnsureType>()) {
             int numStructureType = 0;
             StructureType::Ptr structureType;
 
@@ -618,9 +594,9 @@ void ExpressionVisitor::visitVariableProperty(VariablePropertyAst *node)
             }
         }
 
-        if (type && StructureType::Ptr::dynamicCast(type)) {
+        if (type && type.dynamicCast<StructureType>()) {
             DUChainReadLocker lock(DUChain::lock());
-            Declaration* declaration = StructureType::Ptr::staticCast(type)->declaration(m_currentContext->topContext());
+            Declaration* declaration = type.staticCast<StructureType>()->declaration(m_currentContext->topContext());
             if (declaration) {
                 ifDebug(qCDebug(DUCHAIN) << "parent:" << declaration->toString();)
                 DUContext* context = declaration->internalContext();
@@ -697,20 +673,32 @@ void ExpressionVisitor::visitStaticMember(StaticMemberAst* node)
 {
     //don't call DefaultVisitor::visitStaticMember(node);
     //because we would end up in visitCompoundVariableWithSimpleIndirectReference
-    if (node->staticProperty->staticProperty->variable->variable) {
-        DUContext* context = findClassContext(node->className);
-        if (context) {
-            useDeclaration(node->staticProperty->staticProperty->variable, context);
-        } else {
-            usingDeclaration(node->className, DeclarationPointer());
+    if (node->staticProperty && node->staticProperty->staticProperty) {
+        if (node->staticProperty->staticProperty->variable) {
+            DUContext* context = findClassContext(node->className);
+            if (context) {
+                useDeclaration(node->staticProperty->staticProperty->variable, context);
+            } else {
+                usingDeclaration(node->className, DeclarationPointer());
+                m_result.setType(AbstractType::Ptr());
+            }
+        } else if (node->staticProperty->staticProperty->expr) {
+            const QualifiedIdentifier id = identifierForNamespace(node->className, m_editor);
+            DeclarationPointer declaration = findDeclarationImport(ClassDeclarationType, id);
+            usingDeclaration(node->className->namespaceNameSequence->back()->element, declaration);
+            buildNamespaceUses(node->className, id);
+
+            visitExpr(node->staticProperty->staticProperty->expr);
+
             m_result.setType(AbstractType::Ptr());
         }
-        if (node->staticProperty->offsetItemsSequence) {
-            const KDevPG::ListNode< DimListItemAst* >* it = node->staticProperty->offsetItemsSequence->front();
-            do {
-                visitDimListItem(it->element);
-            } while(it->hasNext() && (it = it->next));
-        }
+    }
+
+    if (node->staticProperty && node->staticProperty->offsetItemsSequence) {
+        const KDevPG::ListNode< DimListItemAst* >* it = node->staticProperty->offsetItemsSequence->front();
+        do {
+            visitDimListItem(it->element);
+        } while(it->hasNext() && (it = it->next));
     }
 }
 
@@ -719,11 +707,18 @@ void ExpressionVisitor::visitClassNameReference(ClassNameReferenceAst* node)
     if (node->staticProperty) {
         DUContext* context = findClassContext(node->className->identifier);
 
-        if (context) {
-            useDeclaration(node->staticProperty->staticProperty->variable, context);
+        if (context && node->staticProperty && node->staticProperty->staticProperty) {
+            if (node->staticProperty->staticProperty->variable) {
+                // static properties (object::$property)
+                useDeclaration(node->staticProperty->staticProperty->variable, context);
+            } else if (node->staticProperty->staticProperty->expr) {
+                // variable static properties (object::${$property})
+                visitExpr(node->staticProperty->staticProperty->expr);
+                usingDeclaration(node->className, DeclarationPointer());
+            }
         }
 
-        if (node->staticProperty->offsetItemsSequence) {
+        if (node->staticProperty && node->staticProperty->offsetItemsSequence) {
             const KDevPG::ListNode< DimListItemAst* >* dim_it = node->staticProperty->offsetItemsSequence->front();
             do {
                 visitDimListItem(dim_it->element);
@@ -749,24 +744,24 @@ void ExpressionVisitor::visitClassNameReference(ClassNameReferenceAst* node)
                         && it->element->property->variableWithoutObjects->variable->variable) {
                     VariableIdentifierAst *varnode = it->element->property->variableWithoutObjects->variable->variable;
                     useDeclaration(varnode, m_currentContext);
+                } else if (it->element->property && it->element->property->variableWithoutObjects
+                        && it->element->property->variableWithoutObjects->variable->expr) {
+                        // variable dynamic properties ($object->${$property})
+                        visitExpr(it->element->property->variableWithoutObjects->variable->expr);
                 } else if (!m_result.allDeclarations().isEmpty()) {
-                    // handle array indices after normal/static properties ($object->property[$index] // $object::$property[$index])
-                    if (it->element->property && it->element->property->objectDimList && it->element->property->objectDimList->offsetItemsSequence) {
-                        const KDevPG::ListNode< DimListItemAst* >* dim_it = it->element->property->objectDimList->offsetItemsSequence->front();
-                        do {
-                            visitDimListItem(dim_it->element);
-                        } while(dim_it->hasNext() && (dim_it = dim_it->next));
-                    } else if (it->element->staticProperty && it->element->staticProperty->offsetItemsSequence) {
-                        const KDevPG::ListNode< DimListItemAst* >* dim_it = it->element->staticProperty->offsetItemsSequence->front();
-                        do {
-                            visitDimListItem(dim_it->element);
-                        } while(dim_it->hasNext() && (dim_it = dim_it->next));
+                    // Handle dynamic static properties first, as they don't need a class context
+                    if (it->element->staticProperty && it->element->staticProperty->staticProperty
+                      && it->element->staticProperty->staticProperty->expr) {
+                        // variable static properties ($object::${$property})
+                        visitExpr(it->element->staticProperty->staticProperty->expr);
+                        usingDeclaration(it->element->staticProperty, DeclarationPointer());
                     }
 
                     type = m_result.allDeclarations().last()->type<StructureType>();
 
                     if (!type) {
                         context = nullptr;
+                        visitClassNameReferenceDimListItems(it->element);
                         continue;
                     }
 
@@ -776,6 +771,7 @@ void ExpressionVisitor::visitClassNameReference(ClassNameReferenceAst* node)
 
                     if (!declaration) {
                         context = nullptr;
+                        visitClassNameReferenceDimListItems(it->element);
                         continue;
                     }
 
@@ -783,10 +779,12 @@ void ExpressionVisitor::visitClassNameReference(ClassNameReferenceAst* node)
 
                     if (!context || context->type() != DUContext::Class) {
                         context = nullptr;
+                        visitClassNameReferenceDimListItems(it->element);
                         continue;
                     }
 
-                    if (it->element->staticProperty) {
+                    if (it->element->staticProperty && it->element->staticProperty->staticProperty
+                      && it->element->staticProperty->staticProperty->variable) {
                         // static properties ($object::$property)
                         VariableIdentifierAst *varnode = it->element->staticProperty->staticProperty->variable;
                         useDeclaration(varnode, context);
@@ -798,11 +796,28 @@ void ExpressionVisitor::visitClassNameReference(ClassNameReferenceAst* node)
                     } else {
                         context = nullptr;
                     }
+
+                    visitClassNameReferenceDimListItems(it->element);
                 }
             } while(it->hasNext() && (it = it->next));
         }
     }
+}
 
+void ExpressionVisitor::visitClassNameReferenceDimListItems(ClassPropertyAst* node)
+{
+    // handle array indices after normal/static properties ($object->property[$index] // $object::$property[$index])
+    if (node->property && node->property->objectDimList && node->property->objectDimList->offsetItemsSequence) {
+        const KDevPG::ListNode< DimListItemAst* >* dim_it = node->property->objectDimList->offsetItemsSequence->front();
+        do {
+            visitDimListItem(dim_it->element);
+        } while(dim_it->hasNext() && (dim_it = dim_it->next));
+    } else if (node->staticProperty && node->staticProperty->offsetItemsSequence) {
+        const KDevPG::ListNode< DimListItemAst* >* dim_it = node->staticProperty->offsetItemsSequence->front();
+        do {
+            visitDimListItem(dim_it->element);
+        } while(dim_it->hasNext() && (dim_it = dim_it->next));
+    }
 }
 
 void ExpressionVisitor::visitUnaryExpression(UnaryExpressionAst* node)
@@ -889,7 +904,7 @@ void ExpressionVisitor::visitStatement(StatementAst *node)
     DefaultVisitor::visitStatement(node);
 
     if (node->returnExpr) {
-        FunctionType::Ptr closureType = currentClosureReturnType().cast<FunctionType>();
+        auto closureType = currentClosureReturnType().dynamicCast<FunctionType>();
 
         if (closureType) {
             closureType->setReturnType(m_result.type());
@@ -897,6 +912,17 @@ void ExpressionVisitor::visitStatement(StatementAst *node)
     }
 }
 
+void ExpressionVisitor::visitGenericTypeHint(GenericTypeHintAst *node) {
+    if (node->genericType && isGenericClassTypehint(node->genericType, m_editor)) {
+        NamespacedIdentifierAst* objectType = node->genericType;
+        QualifiedIdentifier id = identifierForNamespace(objectType, m_editor);
+        DeclarationPointer dec = findDeclarationImport(ClassDeclarationType, id);
+
+        usingDeclaration(objectType->namespaceNameSequence->back()->element, dec);
+        buildNamespaceUses(objectType, id);
+    }
+}
+
 QString ExpressionVisitor::stringForNode(AstNode* id)
 {
     if (!id)
